;; Generated package description from rcirc-mentions.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "rcirc-mentions" "1.0.5.0.20250908.131006" "Log mentions of your nick or keywords in a separate buffer" '((emacs "29.1")) :commit "f48c86345110433cfcda669d3ad6500334371a34" :authors '(("Tassilo Horn" . "tsdh@gnu.org")) :maintainer '("Tassilo Horn" . "tsdh@gnu.org") :keywords '("rcirc" "irc") :url "https://sr.ht/~tsdh/rcirc-mentions/")
